import 'dart:io';

import 'package:antlr4/antlr4.dart';
import 'package:path/path.dart';
import 'package:protobuf3_parser/__generated__/Protobuf3Lexer.dart';
import 'package:protobuf3_parser/__generated__/Protobuf3Parser.dart' as $p;
import 'package:protobuf3_parser/protobuf3_parser.dart';
import 'package:protobuf3_parser/src/string_asset_reader.dart';

import 'const.dart';

class Protobuf3Parser {
  final AssetId _assetId;
  final AssetReader _reader;
  Protobuf3Parser(this._assetId, this._reader);

  static Protobuf3Parser fromString(String string) {
    final assetId = AssetId("");
    final assetReader = StringAssetReader(string);
    return Protobuf3Parser(assetId, assetReader);
  }

  static Protobuf3Parser fromFile(File file) {
    final assetId = AssetId(basename(file.path));
    final assetReader = DirectoryAssetReader(file.parent);
    return Protobuf3Parser(assetId, assetReader);
  }

  // =======================================================
  // == API
  // =======================================================

  /// Get all the imported files.
  Stream<Protobuf3Parser> get files async* {
    for (final importStatement in (await _tree).importStatements()) {
      final strLit = importStatement.strLit()?.STR_LIT()?.text;

      final String path;
      if (strLit != null) {
        path = strLit.substring(1, strLit.length - 1);
      } else {
        throw Exception(
            "Unable to solve the import: '${importStatement.text}'!");
      }
      final assetId = AssetId(path);

      if (_reader.canRead(assetId)) {
        yield Protobuf3Parser(assetId, _reader);
      } else {
        throw Exception("Couldn't import '$path'!");
      }
    }
  }

  /// Get all the [Message3]s in this file.
  Stream<Message3> get messages async* {
    await for (final def in _getTopLevelMessageDefs()) {
      final message = await _getMessage(def);
      if (message != null) yield message;
    }
  }

  /// Get all the [Enum3]s in this file.
  Stream<Enum3> get enums async* {
    await for (final def in _getTopLevelEnumDefs()) {
      final enum3 = _getEnum(def);
      if (enum3 != null) yield enum3;
    }
  }

  /// Get all the [Service3]s in this file.
  Stream<Service3> get services async* {
    await for (final def in _getTopLevelServiceDefs()) {
      final service = await _getService(def);
      if (service != null) yield service;
    }
  }

  /// Get all the [Element3]s in this file.
  Stream<Element3> get elements async* {
    yield* messages;
    yield* enums;
    yield* services;
  }

  /// Get all the [Message3]s from all files we can access.
  Stream<Message3> get allMessages => _getAllMessages();
  Stream<Message3> _getAllMessages([List<AssetId>? readedAssets]) async* {
    readedAssets ??= [];

    // get this file's messages
    yield* messages;
    readedAssets.add(_assetId);

    // get other's messages
    await for (final file in files) {
      if (!readedAssets.contains(file._assetId)) {
        yield* file._getAllMessages(readedAssets..add(file._assetId));
      }
    }
  }

  /// Get all the [Enum]s from all files we can access.
  Stream<Enum3> get allEnums => _getAllEnums();
  Stream<Enum3> _getAllEnums([List<AssetId>? readedAssets]) async* {
    readedAssets ??= [];

    // get this file's messages
    yield* enums;
    readedAssets.add(_assetId);

    // get other's messages
    await for (final file in files) {
      if (!readedAssets.contains(file._assetId)) {
        yield* file._getAllEnums(readedAssets..add(file._assetId));
      }
    }
  }

  /// Get all the [Element3]s from all files we can access.
  Stream<Element3> get allElements => _getAllElements();
  Stream<Element3> _getAllElements([List<AssetId>? readedAssets]) async* {
    readedAssets ??= [];

    // get this file's messages
    yield* elements;
    readedAssets.add(_assetId);

    // get other's messages
    await for (final file in files) {
      if (!readedAssets.contains(file._assetId)) {
        yield* file._getAllElements(readedAssets..add(file._assetId));
      }
    }
  }

  // =======================================================
  // == CONVERTERS
  // =======================================================

  Future<Message3?> _getMessage($p.MessageDefContext ctx) async {
    final messageName = ctx.messageName()?.text;
    final messageBody = ctx.messageBody();
    if (messageName == null || messageBody == null) return null;

    final messageFields = <MessageField>[];

    final messageElements = messageBody.messageElements();
    for (final messageElement in messageElements) {
      final messageField = messageElement.field();
      final isMessageField = messageField != null;

      int iField = 0;
      if (isMessageField) {
        final fieldName = messageField.fieldName()?.text;
        var fieldIdStr = messageField.fieldNumber()?.text;

        final isList = messageField.REPEATED() != null;

        // find the correct id
        final int fieldId;
        if (fieldIdStr == null) {
          fieldId = ++iField;
        } else {
          try {
            fieldId = int.parse(fieldIdStr);
          } catch (_) {
            throw Exception(
                "The field number '$fieldIdStr' isn't a valid integer!");
          }
        }

        // find the correct type
        final fieldTypeStr = messageField.type_()?.text;

        if (fieldTypeStr == null) {
          throw Exception(
              "No type found for the field '$fieldName' in the message '$messageName'!");
        }

        Element3? fieldType = await _getMessageFieldType(ctx, fieldTypeStr);
        if (fieldType == null) {
          throw Exception(
              "Couldn't find definition for the type '$fieldTypeStr'!");
        }

        messageFields.add(MessageField(
          fieldName!,
          fieldType,
          id: fieldId,
          comments: _getComments(messageElement.comments()),
          isList: isList,
        ));
      }
    }

    final message = Message3(messageName);
    message.fields = messageFields.map((e) => e..parent = message).toList();
    message.comments = _getComments(ctx.comments());
    return message;
  }

  Future<Element3?> _getMessageFieldType(
      $p.MessageDefContext ctx, String name) async {
    Element3? fieldType;

    // 1. search in the same level
    await for (final x in _getMessageElements(ctx, false)) {
      if (x is $p.MessageDefContext) {
        if (x.messageName()?.text == name) {
          fieldType = (await _getMessage(x))!;
        }
      } else if (x is $p.EnumDefContext) {
        if (x.enumName()?.text == name) {
          print(x.enumName()?.text);
          fieldType = _getEnum(x);
        }
      }
    }
    if (fieldType != null) return fieldType;

    // 2. search in the top level of all files
    fieldType = await _getMessageFieldTypeFromTopLevel(name);
    if (fieldType != null) return fieldType;

    // 3. check if it's a core type?
    final isCoreType = CoreTypes.values.contains(name);

    if (isCoreType) {
      return CoreTypeMessage(name);
    }

    return null;
  }

  Future<Element3?> _getMessageFieldTypeFromTopLevel(String name,
      [List<AssetId>? readedAssets]) async {
    readedAssets ??= [];

    // 1. search in the top level file
    await for (final x in _getTopLevelMessageDefs()) {
      if (x.messageName()?.text == name) {
        return (await _getMessage(x))!;
      }
    }
    await for (final x in _getTopLevelEnumDefs()) {
      if (x.enumName()?.text == name) {
        return _getEnum(x)!;
      }
    }

    // 2. search other files and it's subfiles...
    // by default, the official protobuf compiler doesn't seems to support loop imports,
    // however I will let it enabled here.
    await for (final file in files) {
      if (readedAssets.contains(file._assetId)) continue;

      final x = await file._getMessageFieldTypeFromTopLevel(
          name, readedAssets..add(file._assetId));
      if (x != null) {
        return x;
      }
    }

    return null;
  }

  Future<Message3?> _getMethodTypeFromTopLevel(String name) async {
    // 1. search in the top level file
    await for (final x in _getTopLevelMessageDefs()) {
      if (x.messageName()?.text == name) {
        return (await _getMessage(x))!;
      }
    }

    // 2. search other files and it's subfiles...
    await for (final file in files) {
      final x = await file._getMethodTypeFromTopLevel(name);
      if (x != null) {
        return x;
      }
    }

    return null;
  }

  List<Comment> _getComments(List<$p.CommentContext> ctxs) => ctxs
      .map((x) => _getComment(x))
      .where((x) => x != null)
      .map((e) => e!)
      .toList();

  Comment? _getComment($p.CommentContext ctx) {
    final line = ctx.LINE_COMMENT()?.text;
    final star = ctx.STAR_COMMENT()?.text;
    final docLine = ctx.DOC_LINE_COMMENT()?.text;

    if (line != null) {
      return Comment(CommentType.single, line);
    } else if (star != null) {
      return Comment(CommentType.multi, star);
    } else if (docLine != null) {
      return Comment(CommentType.singleDoc, docLine);
    } else {
      return null;
    }
  }

  Enum3? _getEnum($p.EnumDefContext ctx) {
    final enumName = ctx.enumName()?.text;
    final enumBody = ctx.enumBody();
    if (enumBody == null) return null;

    final enumFields = <EnumField>[];

    int iField = 0;
    for (final enumElement in enumBody.enumElements()) {
      final enumField = enumElement.enumField();
      if (enumField == null) continue;

      final fieldName = enumField.ident()?.text;
      final fieldIdStr = enumField.intLit()?.text;

      if (fieldName == null) {
        throw Exception("No field name for the enum '$enumName'!");
      }

      final int fieldId;
      if (fieldIdStr == null) {
        fieldId = ++iField;
      } else {
        try {
          fieldId = int.parse(fieldIdStr);
        } catch (_) {
          throw Exception("'$fieldIdStr' is invalid enum id for $enumName!");
        }
      }

      enumFields.add(EnumField(fieldName,
          id: fieldId, comments: _getComments(enumField.comments())));
    }

    if (enumName != null) {
      final enum3 = Enum3(enumName);
      enum3.fields = enumFields.map((e) => e..parent = enum3).toList();
      return enum3;
    } else {
      return null;
    }
  }

  Future<Service3?> _getService($p.ServiceDefContext ctx) async {
    final serviceName = ctx.serviceName()?.text;
    final serviceComments = _getComments(ctx.comments());
    final serviceMethods = <ServiceMethod>[];

    if (serviceName == null) {
      throw Exception("No name found found the service!");
    }

    for (final serviceElement in ctx.serviceElements()) {
      final rpcElement = serviceElement.rpc();
      final isRpcElement = rpcElement != null;

      if (isRpcElement) {
        final rpcName = rpcElement.rpcName()?.text;
        final rpcComments = _getComments(rpcElement.comments());

        if (rpcName == null) {
          throw Exception(
              "No name for the rpc method in the service '$serviceName'!");
        }

        final isStreamRequest = rpcElement.requestStream() != null;
        final isStreamResponse = rpcElement.responseStream() != null;

        final rpcRequestStr = rpcElement.messageType(0)?.text;
        final rpcResponseStr = rpcElement.messageType(1)?.text;

        if (rpcElement.messageTypes().length != 2) {
          throw Exception(
              "Request or response message is missing for the method '$rpcName' in the service '$serviceName'!");
        }

        if (rpcRequestStr == null) {
          throw Exception(
              "The request type '$rpcRequestStr' for the method '$rpcName' in the service '$serviceName' isn't valid!");
        }
        if (rpcResponseStr == null) {
          throw Exception(
              "The response type '$rpcResponseStr' for the method '$rpcName' in the service '$serviceName' isn't valid!");
        }

        final rpcRequest = await _getMethodTypeFromTopLevel(rpcRequestStr);

        if (rpcRequest == null) {
          throw Exception(
              "Couldn't find defintion for the type '$rpcRequestStr'");
        }
        final rpcResponse = await _getMethodTypeFromTopLevel(rpcRequestStr);

        if (rpcResponse == null) {
          throw Exception(
              "Couldn't find defintion for the type '$rpcRequestStr'");
        }

        serviceMethods.add(ServiceMethod(
          rpcName,
          rpcRequest,
          rpcResponse,
          isStreamRequest: isStreamRequest,
          isStreamResponse: isStreamResponse,
          comments: rpcComments,
        ));
      }
    }

    return Service3(serviceName,
        comments: serviceComments, methods: serviceMethods);
  }

  // =======================================================
  // == UTILITES
  // =======================================================

  final _assetsBytes = <int>[];

  Future<$p.ProtoContext> get _tree async {
    if (_assetsBytes.isEmpty) {
      await for (final chunk in _reader.read(_assetId)) {
        _assetsBytes.addAll(chunk);
      }
    }
    final input = InputStream(_assetsBytes);
    final lexer = Protobuf3Lexer(input);
    final tokens = CommonTokenStream(lexer);
    final parser = $p.Protobuf3Parser(tokens);
    final proto = parser.proto();
    return proto;
  }

  Stream<$p.TopLevelDefContext> _getTopLevelDefs() async* {
    final tree = await _tree;
    for (final def in tree.topLevelDefs()) {
      yield def;
    }
  }

  Stream<$p.MessageDefContext> _getTopLevelMessageDefs() async* {
    await for (final def in _getTopLevelDefs()) {
      final messageDef = def.messageDef();
      if (messageDef != null) {
        yield messageDef;
      }
    }
  }

  Stream<$p.EnumDefContext> _getTopLevelEnumDefs() async* {
    await for (final def in _getTopLevelDefs()) {
      final enumDef = def.enumDef();
      if (enumDef != null) {
        yield enumDef;
      }
    }
  }

  Stream<$p.ServiceDefContext> _getTopLevelServiceDefs() async* {
    await for (final def in _getTopLevelDefs()) {
      final serviceDef = def.serviceDef();
      if (serviceDef != null) {
        yield serviceDef;
      }
    }
  }

  /// Get all the sub [$p.MessageDefContext] in the messags.
  Stream<$p.MessageDefContext> _getMessageMessagesDef($p.MessageDefContext ctx,
      [bool sub = true]) async* {
    final messageBody = ctx.messageBody();
    if (messageBody != null) {
      for (final messageElement in messageBody.messageElements()) {
        final messageDef = messageElement.messageDef();
        if (messageDef != null) {
          yield messageDef;
          if (sub) yield* _getMessageMessagesDef(messageDef);
        }
      }
    }
  }

  /// Get all enum definitions found in a message.
  Stream<$p.EnumDefContext> _getMessageEnumDefs(
      $p.MessageDefContext ctx) async* {
    final messageBody = ctx.messageBody();
    if (messageBody != null) {
      for (final messageElement in messageBody.messageElements()) {
        final enumDef = messageElement.enumDef();
        if (enumDef != null) {
          yield enumDef;
        }
      }
    }
  }

  /// Get all enums and the messages that exists under the message.
  Stream<ParserRuleContext> _getMessageElements($p.MessageDefContext ctx,
      [bool sub = true]) async* {
    yield* _getMessageEnumDefs(ctx);
    yield* _getMessageMessagesDef(ctx, sub);

    if (sub) {
      await for (final messageDef in _getMessageMessagesDef(ctx, sub)) {
        yield* _getMessageElements(messageDef, sub);
      }
    }
  }

  @override
  String toString() => "$_assetId";

  @override
  bool operator ==(Object? other) {
    if (other is Protobuf3Parser) {
      return other._assetId == _assetId;
    } else {
      return false;
    }
  }

  @override
  int get hashCode => _assetId.hashCode;
}
